%
% This function implements a number of feature extraction methods.
% In case of mode=0, it implements the method that is described in [1] for the PCA and Gabor wavelet basis functions (Optionally this can
% also be used for the Sparsenet basis functions).
% In case of mode=1, it implements the method that is described in [1] for the Sparsenet basis functions.
%
%function F = FE (X, image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value, ...
%		          codebook, global_mean, mode=0);
% or
%
%function F = FE (X, image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value, ...
%		          codebook, global_mean, mode=1, noise_var, beta, sigma, tol, maxiter, num_flag);
% or
%
%function F = FE (X, image_width, image_height, region_width, region_height, tile_width, tile_height, padding_value, ...
%		          codebook, global_mean, mode=2, num_components, schedule, n_desired, noise_level, use_dc_component, output);
%
%Input:
%  X                : data (one image per column)
%  image_width      : width of the images
%  image_height     : height   - " -
%  region_width     : width of the min/max regions
%  region_height    : height   - " -
%  tile_width       : width of image patches to be sparsely coded
%  tile_height      : height   - " -
%  padding_value    : color to be used if the image is artificially enlarged
%  codebook         : codebook to be used for sparse coding 
%  global_mean      : global mean of the image patches
%  num_components   : number of non-zero components in the coefficient vectors
%
%------- mode = 0 -------------
%
% Use scalar product between each codebook element and each image tile as feature
%
%------- mode = 1 -------------
% In this case  the method uses the conjugate gradient minimization of the Sparsenet
% target function implemented by Bruno Olshausen in the Sparsenet algorithm
% (https://redwood.berkeley.edu/bruno/sparsenet/). See also [2] and [3].
%
% noise_var  : see sparsenet.m
% beta       : - " -
% sigma      : - " -
% tol        : - " -
% maxiter    : - " -
% num_flag   : - " -
%
%------- mode = 2 -------------
% Obtain features by orthogonal matching pursuit or bag of pursuits
%
%  schedule         : (optional) 
%  n_desired        : (optional)
%  noise_level      : (optional)
%  use_dc_component : (optional)
%  output           : (optional)
%
%Output:
%
% F : extracted features
%
%-----------------------
%[1] Kai Labusch, Erhardt Barth, and Thomas Martinetz. Simple Method for High-Performance Digit Recognition Based on Sparse Coding. 
%    IEEE Transactions on Neural Networks, 19(11):1985-1989, 2008.
%    http://www.inb.uni-luebeck.de/publications/pdfs/LaBaMa08c.pdf
%
%[2] Olshausen BA, Field DJ (1997). Sparse coding with an overcomplete
%    basis set: A strategy employed by V1?  Vision Research, 37, 3311-3325.
%    ftp://redwood.ucdavis.edu/pub/papers/VR.pdf
% 
%[3] Olshausen BA, Field DJ (1996). Emergence of simple-cell receptive
%    field properties by learning a sparse code for natural images.
%    Nature, 381, 607-609.
%    ftp://redwood.ucdavis.edu/pub/papers/nature-paper.pdf
%

