#include "mex.h"
#include "csources/utils.h"
#include "csources/DiskUtils.h"
#include "csources/Data.h"
#include "csources/MXData.h"

#include "csources/FeatureExtraction.h"

#include <cstring>
#include <cmath>

using namespace std;

void mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
	if((nlhs >= 2) && (nrhs  == 6)) {

		const mxArray* mx_X	                 = prhs[0];
		const mxArray* mx_image_width            = prhs[1];
		const mxArray* mx_image_height           = prhs[2];
		const mxArray* mx_tile_width             = prhs[3];
		const mxArray* mx_tile_height            = prhs[4];
		const mxArray* mx_number_of_random_tiles = prhs[5];

		int image_width  = (int) *mxGetPr(mx_image_width);
		int image_height = (int) *mxGetPr(mx_image_height);

		int tile_width  = (int) *mxGetPr(mx_tile_width);
		int tile_height = (int) *mxGetPr(mx_tile_height);

		int number_of_random_tiles = (int) *mxGetPr(mx_number_of_random_tiles);
	
		int save_memory=0;
		DataDesc* X = MXDataDesc::Create(mx_X, save_memory);

		if (X->ssize() != image_width*image_height) {

			delete X;

			PRINTF("sample size does not equal image_width*image_height!!!\n");
			FLUSH;

			return;
		}

		RandomTileExtraction* fe_extr = new RandomTileExtraction(image_width, image_height, tile_width, tile_height, number_of_random_tiles);

		DataDesc* features = fe_extr->Extract(X);
		int num_features   = features->ssize();

		double* global_mean = double_array(num_features);
		fe_extr->ComputeGlobalMean(global_mean, features);

		double** FE = NULL;
		features->GetAsMatrix(&FE);

		delete features;

		mxArray* mx_FE = mxCreateNumericMatrix (num_features, number_of_random_tiles, mxDOUBLE_CLASS, mxREAL);
		m_to_vec(mxGetPr(mx_FE), FE, number_of_random_tiles, num_features);

		mxArray* mx_global_mean = mxCreateNumericMatrix (1, num_features, mxDOUBLE_CLASS, mxREAL);
		memcpy(mxGetPr(mx_global_mean), global_mean, num_features*sizeof(double));

		plhs[0] = mx_FE;
		plhs[1] = mx_global_mean;

		delete X;
		delete fe_extr;
		m_free(FE);

		free_array(global_mean);
	}
	else mexPrintf("Wrong number of input/output parameters (%i/%i). \n", nrhs, nlhs);
}

