#include "mex.h"
#include "csources/utils.h"
#include "csources/Data.h"
#include "csources/MXData.h"
#include "csources/FeatureExtraction.h"

#include <cstring>
#include <cmath>

using namespace std;

void mexFunction( int nlhs, mxArray *plhs[],
             	  int nrhs, const mxArray *prhs[] )
{
	const mxArray* mx_X               = prhs [0]; 
	const mxArray* mx_image_size      = prhs [1];
	const mxArray* mx_tile_size       = prhs [2]; 
	const mxArray* mx_number_of_tiles = prhs [3];

	int tile_width = (int) *mxGetPr(mx_tile_size);
	int tile_height = tile_width;

	int image_width = (int) *mxGetPr(mx_image_size);
	int image_height = image_width;
		
	int number_of_tiles = (int) *mxGetPr(mx_number_of_tiles);

	int save_memory = 0;
	DataDesc* X = MXDataDesc::Create(mx_X, save_memory);

	FeatureExtraction* extr = new RandomTileExtraction(image_width, image_height, tile_width, tile_height, number_of_tiles);

	DataDesc* result = extr->Extract(X);
	
	mxArray* mx_result = mxCreateNumericMatrix (result->ssize(), result->size(), mxDOUBLE_CLASS, mxREAL);

	double* dest = mxGetPr(mx_result);

	for (long i=0;i < result->size(); i++) {
		double* cur_tile=NULL;
		result->GetSample(&cur_tile, i);
		memcpy(dest, cur_tile, result->ssize()*sizeof(double));	

		dest += result->ssize();
	}
	plhs[0] = mx_result;

	delete X;
	delete result;
}
