#include "mex.h"
#include "csources/utils.h"
#include "csources/DiskUtils.h"
#include "csources/Data.h"
#include "csources/MXData.h"
#include "csources/SPcgf.h"

#include <cstring>

using namespace std;

void mexFunction( int nlhs, mxArray *plhs[],
             int nrhs, const mxArray *prhs[] )
{
	if((nlhs >= 1) && (nrhs  >= 3)) {

		const mxArray* mx_X	         = prhs[0];
		const mxArray* mx_codebook       = prhs[1];
		const mxArray* mx_noise_var      = prhs[2];
		const mxArray* mx_beta           = prhs[3];
		const mxArray* mx_sigma          = prhs[4];
		const mxArray* mx_maxiter        = prhs[5];
		const mxArray* mx_tol            = prhs[6];

		mxClassID id = mxGetClassID (mx_X);
	
		int save_memory=0;
		DataDesc* X = MXDataDesc::Create(mx_X, save_memory); 

		int num_codebooks = 1;
		int codebook_size = 0;
		int ndims         = mxGetNumberOfDimensions(mx_codebook);		

		if (ndims == 2) codebook_size  = mxGetN(mx_codebook);
		else {
			int* dims = int_array(ndims);
			memcpy(dims, mxGetDimensions(mx_codebook), ndims*sizeof(int));
			codebook_size = dims[1];
			num_codebooks = dims[2];
			free_array(dims);
		}

		if ((num_codebooks > 1) && (num_codebooks < X->size())) {
			PRINTF("Number of codebooks does not match number of samples ... using only the first one ...!!!\n");
			num_codebooks=1;
		}

		double noise_var = (mx_noise_var != NULL)  ? ((!mxIsEmpty(mx_noise_var)) ?  *mxGetPr(mx_noise_var)     : 0.1)    : 0.1;
		double beta      = (mx_beta != NULL)       ? ((!mxIsEmpty(mx_beta))      ?  *mxGetPr(mx_beta)          : 1.0)    : 1.0;
		double sigma     = (mx_sigma != NULL)      ? ((!mxIsEmpty(mx_sigma))     ?  *mxGetPr(mx_sigma)         : 0.316)  : 0.316;
		double tol       = (mx_tol     != NULL)    ? ((!mxIsEmpty(mx_tol))       ?  *mxGetPr(mx_tol)           : 0.001)  : 0.001;
		int maxiter      = (mx_maxiter != NULL)    ? ((!mxIsEmpty(mx_maxiter))   ? (int)  *mxGetPr(mx_maxiter) : 100)    : 100;

		double*** codebooks = (double***) pointer_array(num_codebooks);
		double** pcodebook  = (double**) pointer_array(codebook_size);

		for (int j=0; j < num_codebooks; j++) {

			pcodebook[0] = mxGetPr(mx_codebook) + j*X->ssize()*codebook_size;
			for(int i=0; i < codebook_size; i++) pcodebook[i] = pcodebook[0] + i*X->ssize();

			codebooks [j] = m_create(codebook_size, X->ssize());
			m_copy(codebooks[j], pcodebook, codebook_size, X->ssize());
		}
		free_array(pcodebook);

		double** P = m_create(X->size(), codebook_size);

		int outflag   = 0;
		int numflag=0;	
		double lambda = 1.0/noise_var;
		SPcgf* my_cgf = new SPcgf(codebooks[0], X->ssize(), codebook_size, lambda, beta, sigma, outflag);

		my_cgf->cgf(P, X, tol, maxiter, numflag);	
		delete my_cgf;

		for(int i=0; i < num_codebooks;i++) m_free(codebooks[i]);
		free_array(codebooks);

		mxArray* mx_P = mxCreateNumericMatrix(codebook_size, X->size(), mxDOUBLE_CLASS, mxREAL);
		m_to_vec(mxGetPr(mx_P), P, X->size(), codebook_size);
		m_free(P);

		delete X;

		plhs [0] = mx_P;
	}
	else mexPrintf("Wrong number of input/output parameters (%i/%i). \n", nrhs, nlhs);
}

