#include <cstdio>
#include <cmath>
#include <cstdlib>
#include <cstring>

using namespace std;

#include "utils.h"
#include "ohl.h"
#include "Data.h"

void bop_rec (double** P, double** R, int k, double* cur_r, double noise_level, 
              double** cur_codebook, double* cur_norm, int ssize, int codebook_size, int* cur_candidates, int cur_cand_len, int* used,
              int* schedule, int* n_SOLUTIONS, int n_combis, int** NEXT_WIN, int**** CANDIDATES, int**** CAND_LEN, double***** CODEBOOKS, double**** NORM_CB, double**** RESIDUAL, 
	      double**** Y, double**** P_TO_WIN, int**** Y_ORDER, int**** Y_POS, int**** Y_LEN,  
	      double** NR, int* DEPTH, int* SKIP, int** USE_DEPTH, int* target_combi, int* target_depth, int depth, int use_dc_component, int nonnegative);
/*
	r_P : coefficients 
	r_R : residuals
*/
int bop(double**** r_P, double**** r_R, 
	double** C, int num_components, double noise_level, DataDesc* data, int codebook_size, int* schedule, int n_desired, int use_dc_component, int nonnegative, int output) {

  	if (num_components > codebook_size) {
	  	PRINTF("num_components > codebook_size!!! setting num_components=%i !!!", codebook_size);
		num_components = codebook_size;
	}

	int k=0;

	int X_columns = data->num_samples();
	int ssize     = data->ssize();

	int my_schedule = 0;
	if (schedule == NULL) {
		my_schedule = 1;
		schedule = int_array(num_components);
		iassign(schedule, 1, num_components);
	}

	int n_combis = n_desired*10;
	
	//--------------------------------

	int* candidates = int_array(codebook_size);
	int* used 	= int_array(codebook_size);

	int** NEXT_WIN  = im_create(n_combis, codebook_size);
	int* DEPTH      = int_array(n_combis);
	int* SKIP	= int_array(n_combis);
	int** USE_DEPTH = im_create(n_combis, codebook_size);

	int* next_win = int_array(codebook_size);
	int* cand = int_array(codebook_size);

	int* remainder_nw   = int_array(codebook_size);
	int* remainder_cand = int_array(codebook_size);

	int* I_next_win  = int_array(codebook_size);
	int* I_next_cand = int_array(codebook_size);

	double**** Y    = (double****)pointer_array(n_combis);
	double** hY     = (double**)  pointer_array(n_combis*codebook_size);

	double**** RESIDUAL = (double****)pointer_array(n_combis);
	double** hRESIDUAL  = (double**)  pointer_array(n_combis*codebook_size);

	double**** P_TO_WIN = (double****) pointer_array(n_combis);
	double** hP_TO_WIN  = (double**) pointer_array(n_combis*codebook_size);

	double***** CODEBOOKS = (double*****) pointer_array(n_combis);
	double*** hCODEBOOKS = (double***) pointer_array(n_combis*codebook_size);

	double**** NORM_CB = (double****) pointer_array(n_combis);
	double** hNORM_CB  = (double**) pointer_array(n_combis*codebook_size);

	int**** Y_ORDER = (int****)   pointer_array(n_combis);
	int** hY_ORDER  = (int**)     pointer_array(n_combis*codebook_size);

	int**** Y_POS   = (int****)   pointer_array(n_combis);
	int** hY_POS    = (int**)     pointer_array(n_combis*codebook_size);

	int**** Y_LEN   = (int****)   pointer_array(n_combis);
	int**  hY_LEN   = (int**)     pointer_array(n_combis*codebook_size);

	int**** CANDIDATES = (int****) pointer_array(n_combis);
	int** hCANDIDATES  = (int**)   pointer_array(n_combis*codebook_size);

	int**** CAND_LEN   = (int****)   pointer_array(n_combis);
	int**  hCAND_LEN   = (int**)     pointer_array(n_combis*codebook_size);

	double** NR       = m_create(n_combis, codebook_size);
	int*    max_depth = int_array(n_combis);
	double* max_y     = double_array(n_combis);

	double* ytemp = double_array(codebook_size);
	int*    dtemp = int_array(codebook_size);

	memset(hY        , 0, n_combis*codebook_size*sizeof(double**));
	memset(hY_ORDER  , 0, n_combis*codebook_size*sizeof(int**));
	memset(hY_POS    , 0, n_combis*codebook_size*sizeof(int**));
	memset(hY_LEN    , 0, n_combis*codebook_size*sizeof(int**));
	memset(hCAND_LEN , 0, n_combis*codebook_size*sizeof(int**));

	memset(hP_TO_WIN , 0, n_combis*codebook_size*sizeof(double**));
	memset(hCODEBOOKS, 0, n_combis*codebook_size*sizeof(double***));
	memset(hNORM_CB  , 0, n_combis*codebook_size*sizeof(double**));
	memset(hRESIDUAL , 0, n_combis*codebook_size*sizeof(double**));
	memset(hCANDIDATES, 0, n_combis*codebook_size*sizeof(int**));

	k=0;		
	for (int j=0;j < n_combis; j++) {

		Y[j]         = (double***) pointer_array(codebook_size);
		Y_ORDER[j]   = (int***)    pointer_array(codebook_size);
		Y_POS[j]     = (int***)pointer_array(codebook_size);
		Y_LEN[j]     = (int***)pointer_array(codebook_size);
		CAND_LEN[j]  = (int***)pointer_array(codebook_size);

		P_TO_WIN[j]  = (double***) pointer_array(codebook_size);
		CODEBOOKS[j] = (double****) pointer_array(codebook_size);
		NORM_CB[j]   = (double***) pointer_array(codebook_size);
		RESIDUAL[j]  = (double***) pointer_array(codebook_size);
		CANDIDATES[j] = (int***)pointer_array(codebook_size);

		for (int i=0; i < codebook_size; i++) {
			Y        [j][i]  = hY         + k;
			Y_ORDER  [j][i]  = hY_ORDER   + k;
			Y_POS    [j][i]  = hY_POS     + k;
			Y_LEN    [j][i]  = hY_LEN     + k;

			P_TO_WIN [j][i]  = hP_TO_WIN  + k;
			CODEBOOKS[j][i]  = hCODEBOOKS + k;
			NORM_CB  [j][i]  = hNORM_CB   + k;
			RESIDUAL [j][i]  = hRESIDUAL  + k;
			CANDIDATES[j][i]  = hCANDIDATES + k;
			CAND_LEN [j][i]  = hCAND_LEN  + k;

			k+=1;
		}
	}

	double* temp  = double_array(ssize);
	double* cur_r = double_array(ssize);

	double** C_tmp   = m_create(codebook_size, ssize);
	double* cur_norm = double_array(codebook_size);

	double*** R = (double***) pointer_array(X_columns);
        double*** P = (double***) pointer_array(X_columns);

	double** P_tmp = m_create(n_combis, codebook_size);
	double** R_tmp = m_create(n_combis, ssize);

	double** ptemp = (double**) pointer_array(n_combis);
	int* itemp = int_array(n_combis);

        *r_P = P;
        *r_R = R;

        for (int i=0; i < X_columns;i++) {
                P[i]= m_create(n_desired, codebook_size);
                R[i]= m_create(n_desired, ssize);
        }

	//--------------------------------

	if (output) {
		PRINTF("Performing BOP, num_components: %i, codebook_size: %i, noise level: %g, n_desired: %i\n", num_components, codebook_size, noise_level, n_desired);
		FLUSH;
	}

	double old_percent = 0.0;
	for (int i=0; i < X_columns; i++) {

		double cur_noise_level = noise_level;
		double percent = floor( ((double)i)*100/X_columns);
		if ((old_percent != percent) && output) {
			old_percent = percent;
			PRINTF("%g %% \n", percent);
			FLUSH;
		}

		int* active_dims = NULL;
		int cur_ssize    = data->CopySample(temp, &active_dims, i);

		if (active_dims==NULL) {
			m_copy (C_tmp, C, codebook_size, ssize);
			memcpy(cur_r, temp, ssize*sizeof(double));
		}
		else {
			for(int j=0; j < codebook_size; j++) {
				get_i(C_tmp[j], C[j], active_dims, cur_ssize);
				//double n = sqrt(scp(C_tmp[j], C_tmp[j], cur_ssize));

				//if (n < 10e-13) n = 1.0;
				//sdivide(C_tmp[j], C_tmp[j], n, cur_ssize);
			}
			get_i(cur_r, temp, active_dims, cur_ssize);

			// adjust the noise-level for the missing dimensions
			cur_noise_level *= sqrt(((double) ssize)/((double) cur_ssize));
		}

		int depth        = 0;
		int n_SOLUTIONS  = 0;
		int target_combi = -1;
		int target_depth = -1;

		m_zero(P[i], n_desired, codebook_size);
		for(int j=0; j < n_desired; j++) memcpy(R[i][j], cur_r, cur_ssize*sizeof(double)); // initialize all residuals 

		m_zero(P_tmp, n_combis, codebook_size);
		m_zero(R_tmp, n_combis, ssize);
		for(int j=0; j < n_combis; j++) memcpy(R_tmp[j], cur_r, cur_ssize*sizeof(double)); // initialize all residuals 

		m_zero(NR, n_combis, codebook_size);

		//-------------------------------------------------

		for (int j=0; j < codebook_size; j++) { 
			candidates[j] = j;
			cur_norm[j]   = sqrt(scp(C_tmp[j], C_tmp[j], cur_ssize));
		}

		iassign(max_depth, -1, n_combis); // max_depth controls for which combinations the largest overlap has to be determined
		assign(max_y, -1, n_combis);      // max_y is the largest overlap for each combinations that has been evaluated so far
		
		im_assign(USE_DEPTH, n_combis, codebook_size, codebook_size + 1);
		iassign(SKIP, 1, n_combis);

		do {
			bop_rec (P_tmp, R_tmp, num_components, cur_r, cur_noise_level, 
				C_tmp, cur_norm, cur_ssize, codebook_size, candidates, codebook_size, used,
			        schedule, &n_SOLUTIONS, n_combis, NEXT_WIN, CANDIDATES, CAND_LEN, CODEBOOKS, NORM_CB, RESIDUAL, Y, P_TO_WIN, Y_ORDER, Y_POS, Y_LEN,  
				NR, DEPTH, SKIP, USE_DEPTH, &target_combi, &target_depth, depth, use_dc_component, nonnegative);

			if ((n_SOLUTIONS < n_combis) && (n_SOLUTIONS > 0)) {
				// this code section selects the next combination of codebook vectors that
				// will be replayed up to depth "target_depth" if bop_rec is called again without setting target_combi
				// and target_depth back to -1

				for(int s=0; s < n_SOLUTIONS; s++) { // we iterate through all solutions that have been tried so far

					for(int j=0; j <= DEPTH[s];j++) { // we iterate through all recursion levels of each solution s

						double ** hy = Y[s][j];	       // and pick for each recursion level the next largest overlap with 
						double* y    = *hy;	       // respect to the residual in that level 
			
						int ypos = **(Y_POS[s][j]);

						if (ypos < **(Y_LEN[s][j])) ytemp[j] = fabs(y[ypos])/NR[s][j];
						else ytemp[j]=-1; 		// in this case all possible branches from that recursion level have been tried already
					}
					max_depth[s] = dmax(ytemp, DEPTH[s]+1); // find the largest overlap over all recursion levels for that
					max_y [s]    = ytemp[max_depth[s]];	//  solution
				}

				target_combi = dmax(max_y, n_SOLUTIONS); // select the solution that possess that largest overlap which has not been tested so far
				target_depth = max_depth[target_combi];  // select the recursion level of that solution where the largest overlap occured
			}
			else target_combi = 0; // special case where no solution that satisfies the contraints has been found
			
		} while ((ifind_eq(itemp, SKIP, 0, n_combis) < n_desired) && (n_SOLUTIONS < n_combis) && (max_y[target_combi] > 0));

		//PRINTF("Solutions tested: %i, maximum: %i, max_y[target_combi]: %g\n", n_SOLUTIONS, n_combis, max_y[target_combi]);
		//FLUSH;

		int n_success = ifind_eq(itemp, SKIP, 0, n_combis);
		//PRINTF("n_success: %i\n", n_success);

		n_success = (n_success > n_desired) ? n_desired : n_success;

		pget_i(ptemp, P_tmp, itemp, n_success);	
		m_copy(P[i], ptemp, n_success, codebook_size);
		
		if (active_dims != NULL) {
			m_zero(R[i], n_desired, ssize);
			for(int j=0; j < n_desired; j++) set_i(R[i][j], cur_r, active_dims, cur_ssize);
		}

		pget_i(ptemp, R_tmp, itemp, n_success);
		m_copy(R[i], ptemp, n_success, cur_ssize);

		if (active_dims != NULL) { // if some dimensions have been omitted ... we have to restore the order within the residual
			for (int j=0; j < n_success; j++) {
				memcpy(temp, R[i][j], ssize*sizeof(double));
				memset(R[i][j], 0, ssize*sizeof(double));
				set_i(R[i][j], temp, active_dims, cur_ssize);
			} 
		}

		//--------------------------------------------------
		// free mem that has been allocated by bop_rec during previous calls

		k=0;		
		for (int j=0;j < n_combis; j++) {
			for (int m=0; m < codebook_size; m++) {
				if (*(Y[j][m]) != NULL) {

					free_array(*(Y[j][m]));	
					free_array(*(Y_ORDER[j][m]));
					free_array(*(Y_POS[j][m]));
					free_array(*(Y_LEN[j][m]));
					

					*(Y[j][m])         = NULL;
					*(Y_ORDER[j][m])   = NULL;
					*(Y_POS[j][m])     = NULL;
					*(Y_LEN[j][m])     = NULL;
				}

				if (*(CODEBOOKS[j][m]) != NULL) {
					free_array(*(P_TO_WIN[j][m]));
					m_free(*(CODEBOOKS[j][m]));	
					free_array(*(NORM_CB[j][m]));
					free_array(*(RESIDUAL[j][m]));
					free_array(*(CANDIDATES[j][m]));
					free_array(*(CAND_LEN[j][m]));

					*(P_TO_WIN[j][m])  = NULL;
					*(CODEBOOKS[j][m]) = NULL;
					*(NORM_CB[j][m])   = NULL;
					*(RESIDUAL[j][m])  = NULL;
					*(CANDIDATES[j][m])  = NULL;
					*(CAND_LEN[j][m])   = NULL;
				}

				Y        [j][m] = hY         + k;
				Y_ORDER  [j][m] = hY_ORDER   + k;
				Y_POS    [j][m] = hY_POS     + k;
				Y_LEN    [j][m] = hY_LEN     + k;


				P_TO_WIN [j][m] = hP_TO_WIN  + k;
				CODEBOOKS[j][m] = hCODEBOOKS + k;
				NORM_CB  [j][m] = hNORM_CB   + k;
				RESIDUAL [j][m] = hRESIDUAL  + k;
				CANDIDATES[j][m] = hCANDIDATES + k;
				CAND_LEN  [j][m] = hCAND_LEN     + k;

				k+=1;
			}
		}
	}

	free_array(candidates);
	free_array(used);
	free_array(cur_r);
	free_array(temp);
	free_array(DEPTH);
	free_array(SKIP);
	im_free(NEXT_WIN);
	im_free(USE_DEPTH);

	free_array(next_win);
	free_array(cand);
	free_array(remainder_nw);
	free_array(remainder_cand);
	free_array(I_next_win);
	free_array(I_next_cand);

	for (int j=0;j < n_combis; j++) {
		free_array(Y[j]);
		free_array(Y_ORDER[j]);
		free_array(Y_POS[j]);
		free_array(Y_LEN[j]);

		free_array(P_TO_WIN[j]);
		free_array(CODEBOOKS[j]);
		free_array(NORM_CB[j]);
		free_array(RESIDUAL[j]);
		free_array(CANDIDATES[j]);
		free_array(CAND_LEN[j]);
	}

	free_array(Y);
	free_array(hY);

	free_array(P_TO_WIN);
	free_array(hP_TO_WIN);
	
	free_array(CODEBOOKS);
	free_array(hCODEBOOKS);

	free_array(NORM_CB);
	free_array(hNORM_CB);

	free_array(RESIDUAL);
	free_array(hRESIDUAL);

	free_array(Y_ORDER);
	free_array(hY_ORDER);

	free_array(Y_POS);
	free_array(hY_POS);

	free_array(Y_LEN);
	free_array(hY_LEN);

	free_array(CANDIDATES);
	free_array(hCANDIDATES);

	free_array(CAND_LEN);
	free_array(hCAND_LEN);

	m_free(NR);
	m_free(C_tmp);
	m_free(P_tmp);
	m_free(R_tmp);
	free_array(ptemp);
	free_array(itemp);
	free_array(cur_norm);
	free_array(max_depth);
	free_array(max_y);

	free_array(ytemp);
	free_array(dtemp);

	if (my_schedule) free_array(schedule);

	return n_desired;
}

void bop_rec (double** P, double** R, int k, double* cur_r, double noise_level, 
              double** cur_codebook, double* cur_norm, int ssize, int codebook_size, int* cur_candidates, int cur_cand_len, int* used,
              int* schedule, int* n_SOLUTIONS, int n_combis, int** NEXT_WIN, int**** CANDIDATES, int**** CAND_LEN, double***** CODEBOOKS, double**** NORM_CB, double**** RESIDUAL, 
	      double**** Y, double**** P_TO_WIN, int**** Y_ORDER, int**** Y_POS, int**** Y_LEN,
	      double** NR, int* DEPTH, int* SKIP, int** USE_DEPTH, int* target_combi, int* target_depth, int depth, int use_dc_component, int nonnegative) {

	if (*n_SOLUTIONS >= n_combis) {
 //	PRINTF("!!!!!!!!!!!!!!\n");
		return;
	}

	double* y = NULL;
	int*    I = NULL;
	int* ypos = NULL;
	int* ylen = NULL;

	double** hy = NULL;
	int** hI    = NULL;

	int** hYPOS = NULL;
	int** hYLEN = NULL;

	int l_start = 0;
	int l_step  = schedule[depth];

	if ((*target_combi >= 0) && (depth < *target_depth) && (*(Y[*target_combi][depth]) != NULL)) {  // we replay combination *target_combi up to recursion level *target_depth

		hy        = Y        [*target_combi][depth]; 		    // y and I already have been computed
		hI        = Y_ORDER  [*target_combi][depth];
		hYPOS     = Y_POS    [*target_combi][depth];
		hYLEN     = Y_LEN    [*target_combi][depth];
		
		y        = *hy;						    // overlap between residual and current codebook
		I        = *hI;						    // indices of the codebook vectors that belong to the overlap values
		ypos     = *hYPOS;					    // the overlap values of in y have been tried up to position "*ypos" 
		ylen     = *hYLEN;					    // "*ylen" is the length of y and I

		l_start=0;
		l_step=1;
	}
	else { // otherwise, we are in/enter unknown waters

		if((*target_depth >= 0) && (depth >= *target_depth)) {              // if *target_depth has been reached, we just enter the "try something new" mode

			hy        = Y        [*target_combi][depth]; 		    // y and I already have been computed (we just reached target_depth!)
			hI        = Y_ORDER  [*target_combi][depth];
			hYPOS     = Y_POS    [*target_combi][depth];
			hYLEN     = Y_LEN    [*target_combi][depth];

			y        = *hy;						    // overlap between residual and current codebook
			I        = *hI;						    // indices of the codebook vectors that belong to the overlap values
			ypos     = *hYPOS;					    // the overlap values of in y have been tried up to position "*ypos" 
			ylen     = *hYLEN;					    // "*ylen" is the length of y and I

			l_start = *ypos;
			l_step  = schedule[depth];

			if (l_start + l_step > *ylen)  l_step = *ylen - l_start;

			*ypos += l_step;

			*target_depth = -1;
			*target_combi = -1;
		}
		else {								    // in this case, we are completely in unknown waters and have to compute everything
			hy        = Y          [*n_SOLUTIONS][depth];
			hI        = Y_ORDER    [*n_SOLUTIONS][depth];
			hYPOS     = Y_POS      [*n_SOLUTIONS][depth];
			hYLEN     = Y_LEN      [*n_SOLUTIONS][depth];

			if (!nonnegative) {	
				I          = int_array(cur_cand_len);
				y          = double_array(codebook_size);
				ypos       = int_array(1);
				ylen       = int_array(1);

				*hy        = y;	
				*hI        = I;
				*hYPOS     = ypos;
				*hYLEN     = ylen;

				*ylen   = cur_cand_len;
				*ypos   = 0;
				if ((use_dc_component) && (depth == 0)) *ypos = *ylen-1; 		    // if dc component is used -> only one possibility at depth 0

				l_start = *ypos;
				l_step  = schedule[depth];	

				if (l_start + l_step > *ylen) l_step = *ylen - l_start;

				*ypos += l_step;

				double* temp = double_array(cur_cand_len);
				int * itemp  = int_array(cur_cand_len);

				for (int j=0; j < cur_cand_len; j++) { 			    // calculate distance of the current residual to all codebook vectors, i.e, y = w*x
					y[j] = scp(cur_r, cur_codebook[cur_candidates[j]], ssize);
					itemp[j] = j;
				}
	
				multiply(temp , y   , y , cur_cand_len); 	// temp = -(y^2)
				smultiply(temp, temp, -1, cur_cand_len); 

				isort(temp, itemp, cur_cand_len); 		// sort temp

				memcpy(temp, y, cur_cand_len*sizeof(double));
				get_i (y, temp          , itemp, cur_cand_len); // reorder y according to the sorting
				iget_i(I, cur_candidates, itemp, cur_cand_len); // map sorting to codebook-vector-indices

				free_array(itemp);
				free_array(temp);
			}
			else {
				I          = int_array(cur_cand_len);
				y          = double_array(cur_cand_len);
				ypos       = int_array(1);
				ylen       = int_array(1);

				double* temp = double_array(codebook_size);
				int * itemp  = int_array(codebook_size);

				double* ytemp = double_array(cur_cand_len);
				int* is_nn    = int_array(cur_cand_len);
				
				for (int j=0; j < cur_cand_len; j++)  			    // calculate distance of the current residual to all codebook vectors, i.e, y = w*x
					ytemp[j] = scp(cur_r, cur_codebook[cur_candidates[j]], ssize);

				////////////////////////////////////////
				// This section checks if the selection of candidate w leads to non-negative coefficients

				double* PTEMP = double_array(codebook_size);

				for (int w=0; w < cur_cand_len;w++) {

					memset(PTEMP, 0, codebook_size*sizeof(double));

					double nwin = 1.0;
					if (cur_norm) nwin = cur_norm[cur_candidates[w]];

					PTEMP[cur_candidates[w]] = ytemp[w]/(nwin*nwin);
					used[depth] = cur_candidates[w];

					for (int d1=depth-1; d1 >= 0; d1--) {

						for (int d2=depth; d2 > d1; d2--) {
							PTEMP[used[d2]] = PTEMP[used[d2]]/(*NORM_CB[*n_SOLUTIONS][d1])[used[d2]];

							PTEMP[used[d1]] = PTEMP[used[d1]] - PTEMP[used[d2]]*(*P_TO_WIN[*n_SOLUTIONS][d1])[used[d2]];
						}

						PTEMP[used[d1]] = PTEMP[used[d1]] + P[*n_SOLUTIONS][used[d1]];
					}

					int r = (int) (find_l(itemp, PTEMP, 0, codebook_size) == 0);

					is_nn[w] = r;
				}

				free_array(PTEMP);

				/////////////////////////////////////

				int num_non_negative = ifind_eq(itemp, is_nn, 1, cur_cand_len);

				multiply(temp , ytemp   , ytemp , cur_cand_len); 	// temp = -(ytemp^2)

				get_i(y, temp, itemp, num_non_negative);	// select those that lead to non-negative coefficients

				smultiply(y, y, -1, num_non_negative); 

				isort(y, itemp, num_non_negative); 		// sort temp

				//memcpy(temp, y, num_non_negative*sizeof(double));
				get_i (y, ytemp         , itemp, num_non_negative); // reorder y according to the sorting
				iget_i(I, cur_candidates, itemp, num_non_negative); // map sorting to codebook-vector-indices

				free_array(itemp);
				free_array(temp);
			
				free_array(ytemp);
				free_array(is_nn);

				/////////////////////////////////////

				if (num_non_negative > 0) {

					*ylen   = num_non_negative;
					*ypos   = 0;
					if ((use_dc_component) && (depth == 0)) *ypos = *ylen-1; 		    // if dc component is used -> only one possibility at depth 0

					l_start = *ypos;
					l_step  = schedule[depth];	

					if (l_start + l_step > *ylen) l_step = *ylen - l_start;

					*ypos += l_step;

					*hy        = y;	
					*hI        = I;
					*hYPOS     = ypos;
					*hYLEN     = ylen;
				}
				else  {
					free_array(I);
					free_array(y);
					free_array(ypos);
					free_array(ylen);
					return;
				}
			}
		}
	}

	for (int l=l_start;l < l_start + l_step;l++) {

		double** C       = NULL;
		double* p_to_win = NULL;
		double* n        = NULL;
		double* r 	 = NULL;
		int* candidates  = NULL;

		double*** hC 	   = NULL;
		double** hp_to_win = NULL;
		double** hn 	   = NULL;
		double** hr	   = NULL;
		int** hcand	   = NULL;
		int** hCAND_LEN    = NULL;
		int* cand_len      = NULL;

		int win      = 0;
		double y_win = 0.0;
		double nwin  = 0.0;
		int skip     = 0;
		double nr    = 0.0;

		if (*target_combi == -1) { // if *target_combi == -1 holds, we actually try to select a new combination 

			hp_to_win = P_TO_WIN   [*n_SOLUTIONS][depth];
			hC	  = CODEBOOKS  [*n_SOLUTIONS][depth];
			hn 	  = NORM_CB    [*n_SOLUTIONS][depth];
			hr	  = RESIDUAL   [*n_SOLUTIONS][depth];
			hcand	  = CANDIDATES [*n_SOLUTIONS][depth];
			hCAND_LEN = CAND_LEN   [*n_SOLUTIONS][depth];

			if (!(use_dc_component) || (depth > 0)) { 
				win   = I[l];
				y_win = y[l];
			}
			else {
				ifind_eq(&win, I, 0, *ylen); // if dc component is used -> always use component 0 at depth 0
				y_win = y[win];
				win   = 0;
			}

			if (cur_norm) nwin = cur_norm[win];
			else nwin = 1.0;

			if (nwin < 1e-6) nwin = 1;

			// check if the new combination, that is to be tested already occcured somewhere in different order
			used[depth]=win;

			int num_found = 0;

			int* combi_candidates = int_array(*n_SOLUTIONS);
			iassign(combi_candidates, 1, *n_SOLUTIONS);

			for (int d=0; d <= depth;d++) {
				for(int j=0; j < *n_SOLUTIONS;j++) {
					if (USE_DEPTH[j][used[d]] > depth) combi_candidates[j]=0;
				}
			}

			// now combi_candidates equals to 1 only for those combinations that use exactly the same dictionary
			// elements up to recursion level depth

			num_found = isum(combi_candidates, *n_SOLUTIONS);

			for (int d=0; d <= depth;d++) {
				for(int j=0; j < *n_SOLUTIONS;j++) {
					if((DEPTH[j] >= d) && (combi_candidates[j])){
	
						if ((**(Y_POS[j][d]) < **(Y_LEN[j][d]))) {
							if ((*(Y_ORDER[j][d]))[**(Y_POS[j][d])] == win) { (**(Y_POS[j][d]))++;}
						}
					}
				}
			}

			free_array(combi_candidates);

			if (num_found == 0) {

				if (*hC == NULL) {

					C          = m_create(codebook_size, ssize);		    // this has to be computed anyway
					p_to_win   = double_array(codebook_size);
					n          = double_array(codebook_size);
					r          = double_array(ssize);
					candidates = int_array(*ylen);
				  	cand_len   = int_array(1);	
					
					*cand_len = iremove_value(candidates, I, win, *ylen); // remove winner from list of candidates

					*hp_to_win = p_to_win;
					*hC        = C;
					*hn	   = n;
					*hr 	   = r;
					*hcand 	   = candidates;
					*hCAND_LEN = cand_len;
				}
				else {
					p_to_win = *hp_to_win;
					C        = *hC;
					n	 = *hn;
					r	 = *hr;
					candidates = *hcand;
					cand_len = *hCAND_LEN;

					*cand_len = iremove_value(candidates, cur_candidates, win, *cand_len); // remove winner from list of candidates
				}

				double* temp2 = double_array(ssize);

				m_copy(C, cur_codebook, codebook_size, ssize); 				 // copy current codebook
	
				memset(n       , 0, codebook_size*sizeof(double));
				memset(p_to_win, 0, codebook_size*sizeof(double));

				for (int j=0; j < *cand_len; j++) {					// set norm of remaining candidates to 1, winner remains unchanged!!!

					p_to_win[candidates[j]] = scp(C[candidates[j]], C[win], ssize)/(nwin*nwin); 		// compute projection of codebook to winning codebook_vector

					smultiply(temp2, C[win], p_to_win[candidates[j]], ssize);
					sub(C[candidates[j]], C[candidates[j]], temp2, ssize);					// remove projection from codebook

					n[candidates[j]] = sqrt(scp(C[candidates[j]], C[candidates[j]], ssize));

					if (n[candidates[j]] <= 1e-6) n[candidates[j]] = 1.0;

					sdivide(C[candidates[j]], C[candidates[j]], n[candidates[j]], ssize);
				}

				smultiply(temp2, C[win], y_win/(nwin*nwin), ssize); 				 // remove projection of residual to winner from residual
				sub (r, cur_r, temp2, ssize);

				nr = sqrt(scp(r,r,ssize));

				free_array(temp2);
			}
			else skip = 1;
		}
		else {						// if *target_combi != -1 holds, we are still replaying some solution up to depth "*target_depth"

			hp_to_win = P_TO_WIN  [*target_combi][depth];		    // the projections as well as the orthogonalized codebook can be reused
			hC	  = CODEBOOKS [*target_combi][depth];
			hn        = NORM_CB   [*target_combi][depth];
			hr        = RESIDUAL  [*target_combi][depth];
			hcand     = CANDIDATES[*target_combi][depth];
			hCAND_LEN = CAND_LEN  [*target_combi][depth];

			p_to_win = *hp_to_win;
			C        = *hC;
			n	 = *hn;
			r	 = *hr;
			candidates = *hcand;
			cand_len = *hCAND_LEN;

			ifind_eq(&win, I, NEXT_WIN[*target_combi][DEPTH[*target_combi] - depth], *ylen);
			y_win = y[win];	
			win   = NEXT_WIN[*target_combi][DEPTH[*target_combi] - depth];

			nr = NR[*target_combi][depth+1];

			if (cur_norm) nwin  = cur_norm[win];
			else nwin = 1.0;

			if (nwin < 1e-6) nwin = 1;						 // prevent NaNs
		}

		if ((k > 1) && (nr > noise_level) && (nr > 1e-8) &&(!skip)) {

			int old_n_SOLUTIONS = *n_SOLUTIONS;
			used[depth] = win;

			P        [*n_SOLUTIONS][win] = y_win/(nwin*nwin);  // this section is required in order to be able to perform NN-sparse approximation

			Y        [*n_SOLUTIONS][depth] = hy;
			Y_ORDER  [*n_SOLUTIONS][depth] = hI;
			Y_POS    [*n_SOLUTIONS][depth] = hYPOS;
			Y_LEN    [*n_SOLUTIONS][depth] = hYLEN;
			NR       [*n_SOLUTIONS][depth] = sqrt(scp(cur_r, cur_r, ssize));

			P_TO_WIN [*n_SOLUTIONS][depth] = hp_to_win;
			CODEBOOKS[*n_SOLUTIONS][depth] = hC;
			NORM_CB  [*n_SOLUTIONS][depth] = hn;
			RESIDUAL [*n_SOLUTIONS][depth] = hr;
			CANDIDATES[*n_SOLUTIONS][depth] = hcand;
			CAND_LEN  [*n_SOLUTIONS][depth] = hCAND_LEN;

			USE_DEPTH[*n_SOLUTIONS][win] = depth; 


			bop_rec (P, R, k-1, r, noise_level, 
			          C, NULL, ssize, codebook_size, candidates, *cand_len, used,
			          schedule, n_SOLUTIONS, n_combis, NEXT_WIN, CANDIDATES, CAND_LEN, CODEBOOKS, NORM_CB, RESIDUAL, Y, P_TO_WIN, Y_ORDER, Y_POS, Y_LEN,  
				  NR, DEPTH, SKIP, USE_DEPTH, target_combi, target_depth, depth+1, use_dc_component, nonnegative);

			if (old_n_SOLUTIONS != *n_SOLUTIONS) {

				for (int s=old_n_SOLUTIONS; s < *n_SOLUTIONS; s++) { // we have to normalize and track the results of all solutions that have been obtained from the recursion

					int nw_len = DEPTH[s] - depth;
					P[s][win] = y_win/(nwin*nwin);
	
					for (int j=0; j < nw_len; j++) {
						P[s][win] = P[s][win] - P[s][NEXT_WIN[s][j]]*p_to_win[NEXT_WIN[s][j]]/n[NEXT_WIN[s][j]];
						P[s][NEXT_WIN[s][j]] = P[s][NEXT_WIN[s][j]]/n[NEXT_WIN[s][j]];
					}

					NEXT_WIN [s][nw_len] = win;
				
					Y        [s][depth] = hy;
					Y_ORDER  [s][depth] = hI;
					Y_POS    [s][depth] = hYPOS;
					Y_LEN    [s][depth] = hYLEN;
					NR       [s][depth] = sqrt(scp(cur_r, cur_r, ssize));

					P_TO_WIN [s][depth] = hp_to_win;
					CODEBOOKS[s][depth] = hC;
					NORM_CB  [s][depth] = hn;
					RESIDUAL [s][depth] = hr;
					CANDIDATES[s][depth] = hcand;
					CAND_LEN [s][depth] = hCAND_LEN;

					USE_DEPTH[s][win] = depth; 
				}
			}
			else {
				Y        [*n_SOLUTIONS][depth] = hy;
				Y_ORDER  [*n_SOLUTIONS][depth] = hI;
				Y_POS    [*n_SOLUTIONS][depth] = hYPOS;
				Y_LEN    [*n_SOLUTIONS][depth] = hYLEN;
				NR       [*n_SOLUTIONS][depth] = sqrt(scp(cur_r, cur_r, ssize));

				P_TO_WIN [*n_SOLUTIONS][depth] = hp_to_win;
				CODEBOOKS[*n_SOLUTIONS][depth] = hC;
				NORM_CB  [*n_SOLUTIONS][depth] = hn;
				RESIDUAL [*n_SOLUTIONS][depth] = hr;
				CANDIDATES[*n_SOLUTIONS][depth] = hcand;
				CAND_LEN [*n_SOLUTIONS][depth] = hCAND_LEN;

				DEPTH    [*n_SOLUTIONS]      = depth;	
				P        [*n_SOLUTIONS][win] = y_win/(nwin*nwin);
				NEXT_WIN [*n_SOLUTIONS][0]   = win;

				USE_DEPTH[*n_SOLUTIONS][win] = depth; 

				if (!skip) memcpy(R[*n_SOLUTIONS], r, ssize*sizeof(double));

				SKIP[*n_SOLUTIONS] = skip;

				(*n_SOLUTIONS) += 1;
			}
		}
		else {
			Y        [*n_SOLUTIONS][depth] = hy;
			Y_ORDER  [*n_SOLUTIONS][depth] = hI;
			Y_POS    [*n_SOLUTIONS][depth] = hYPOS;
			Y_LEN    [*n_SOLUTIONS][depth] = hYLEN;
			NR       [*n_SOLUTIONS][depth] = sqrt(scp(cur_r, cur_r, ssize));

			P_TO_WIN [*n_SOLUTIONS][depth] = hp_to_win;
			CODEBOOKS[*n_SOLUTIONS][depth] = hC;
			NORM_CB  [*n_SOLUTIONS][depth] = hn;
			RESIDUAL [*n_SOLUTIONS][depth] = hr;
			CANDIDATES[*n_SOLUTIONS][depth] = hcand;
			CAND_LEN [*n_SOLUTIONS][depth] = hCAND_LEN;

			DEPTH    [*n_SOLUTIONS]      = depth;	
			P        [*n_SOLUTIONS][win] = y_win/(nwin*nwin);
			NEXT_WIN [*n_SOLUTIONS][0]   = win;

			USE_DEPTH[*n_SOLUTIONS][win] = depth; 

			if (!skip) memcpy(R[*n_SOLUTIONS], r, ssize*sizeof(double));

			SKIP[*n_SOLUTIONS] = skip;

			(*n_SOLUTIONS) += 1;
		}
	}
}
